#!/usr/bin/env python
#
# Created by Pearu Peterson in Febuary 2008
#

import os

from macros import preprocess
cwd = os.path.abspath(os.path.dirname(__file__))
targetfile_py = os.path.join(cwd,'..','sympycore','basealgebra','pairs_ops.py')

template = '''\
"""Generated low-level arithmetic methods for CollectingField.

This file is generated by the src/mk_pairs_ops.py script.
See http://sympycore.googlecode.com/ for more information.

DO NOT CHANGE THIS FILE DIRECTLY!!!
"""

from ..core import Expr
from ..utils import NUMBER, SYMBOL, TERMS, FACTORS
from ..arithmetic.numbers import (normalized_fraction, mpq, try_power,
  numbertypes, inttypes_set, numbertypes_set)
from ..arithmetic.infinity import Infinity

def div(a, b, cls):
    if type(b) in inttypes_set:
        if type(a) in inttypes_set:
            if not b:
                if not a:
                    return cls.undefined
                return cls.zoo
            return normalized_fraction(a, b)
        if not b:
            return a * cls.zoo
        if b == 1:
            return a
        return a * mpq((1,b))
    return a / b

'''

#======================================
# General macros
#======================================

RETURN_NEW2 = '''\
%(TMP)s = %(DATA)s
if isinstance(%(TMP)s, Infinity):
    return %(TMP)s
@RETURN_NEW(HEAD=%(HEAD)s; DATA=%(TMP)s)
'''

CANONIZE_TERMS_DICT = '''\
if not %(DICT)s:
    return cls.zero
if len(%(DICT)s)==1:
   t, c = %(DICT)s.items()[0]
   if c==1:
       return t
   if t==cls.one:
       return cls(NUMBER, c)
'''

CANONIZE_FACTORS_DICT = '''\
if not %(DICT)s:
    if %(NUMBER)s is 1:
        return cls.one
    return %(NUMBER)s
if len(%(DICT)s)==1:
   t, c = %(DICT)s.items()[0]
   if c==1:
       if %(NUMBER)s==1:
           return t
       @RETURN_NEW(HEAD=TERMS; DATA={t: %(NUMBER)s})
   if t==cls.one:
       return %(NUMBER)s
'''

CANONIZE_FACTORS_DICT1 = '''\
if not %(DICT)s:
    return cls.one
if len(%(DICT)s)==1:
   t, c = %(DICT)s.items()[0]
   if c==1:
       return t
   if t==cls.one:
       return t
'''

ADD_VALUE_DICT='''\
@ADD_TERM_VALUE_DICT(TERM=cls.one; VALUE=%(VALUE)s; DICT=%(DICT)s; DICT_GET=%(DICT)s.get; SIGN=+; USIGN=)
'''

NEG_DICT_VALUES = '''\
%(DICT_OUT)s = dict([(t, -c) for t,c in %(DICT_IN)s.iteritems()])
'''
MUL_DICT_VALUES = '''\
%(DICT_OUT)s = dict([(t, c*%(OP)s) for t,c in %(DICT_IN)s.iteritems()])
'''

#======================================
# NEG macros
#======================================

NEG_NUMBER = '@RETURN_NEW(HEAD=NUMBER; DATA=-%(OPDATA)s)\n'
NEG_SYMBOL = '@RETURN_NEW(HEAD=TERMS; DATA={%(OP)s: -1})\n'
NEG_TERMS = '''\
op_pairs = %(OPDATA)s
if len(op_pairs)==1:
    t, c = op_pairs.items()[0]
    c = -c
    if c==1:
        return t
    @RETURN_NEW(HEAD=TERMS; DATA={t:c})
@NEG_DICT_VALUES(DICT_IN=%(OPDATA)s; DICT_OUT=pairs)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''
NEG_FACTORS = '@NEG_SYMBOL(OP=%(OP)s; OPDATA=%(OPDATA)s)\n'

#======================================
# ADD macros
#======================================

ADD_VALUE_NUMBER='@RETURN_NEW(HEAD=NUMBER; DATA=%(VALUE)s + %(RHSDATA)s)\n'
ADD_NUMBER_VALUE='@RETURN_NEW(HEAD=NUMBER; DATA=%(LHSDATA)s + %(VALUE)s)\n'
ADD_NUMBER_NUMBER = '@ADD_VALUE_NUMBER(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
ADD_VALUE_SYMBOL = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    return %(RHS)s
@RETURN_NEW(HEAD=TERMS; DATA={cls.one: %(TMP)s, %(RHS)s: 1})
'''
ADD_SYMBOL_VALUE = '@ADD_VALUE_SYMBOL(VALUE=%(VALUE)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
ADD_NUMBER_SYMBOL = '@ADD_VALUE_SYMBOL(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
ADD_SYMBOL_NUMBER = '@ADD_SYMBOL_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
ADD_SYMBOL_SYMBOL = '''\
if %(LHS)s == %(RHS)s:
    pairs = {%(LHS)s: 2}
else:
    pairs = {%(LHS)s: 1, %(RHS)s: 1}
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''
ADD_VALUE_TERMS = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    return %(RHS)s
pairs = %(RHSDATA)s.copy()
@ADD_VALUE_DICT(DICT=pairs; VALUE=%(TMP)s)
@CANONIZE_TERMS_DICT(DICT=pairs)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''
ADD_TERMS_VALUE = '@ADD_VALUE_TERMS(VALUE=%(VALUE)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
ADD_NUMBER_TERMS = '@ADD_VALUE_TERMS(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
ADD_TERMS_NUMBER = '@ADD_TERMS_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'

ADD_TERMS_SYMBOL = '''\
pairs = %(LHSDATA)s.copy()
@ADD_TERM_VALUE_DICT(TERM=%(RHS)s; VALUE=1; DICT=pairs; DICT_GET=pairs.get; SIGN=+; USIGN=)
@CANONIZE_TERMS_DICT(DICT=pairs)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''
ADD_SYMBOL_TERMS = '@ADD_TERMS_SYMBOL(LHS=%(RHS)s; LHSDATA=%(RHSDATA)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
ADD_TERMS_TERMS = '''\
pairs = %(LHSDATA)s.copy()
pairs_get = pairs.get
for t,c in %(RHSDATA)s.iteritems():
    @ADD_TERM_VALUE_DICT(TERM=t; VALUE=c; DICT=pairs; DICT_GET=pairs_get; SIGN=+; USIGN=)
@CANONIZE_TERMS_DICT(DICT=pairs)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''

#======================================
# SUB macros
#======================================

SUB_VALUE_NUMBER='@RETURN_NEW(HEAD=NUMBER; DATA=%(VALUE)s - %(RHSDATA)s)\n'
SUB_NUMBER_VALUE='@RETURN_NEW(HEAD=NUMBER; DATA=%(LHSDATA)s - %(VALUE)s)\n'
SUB_NUMBER_NUMBER = '@SUB_VALUE_NUMBER(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
SUB_VALUE_SYMBOL = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    @RETURN_NEW(HEAD=TERMS; DATA={%(RHS)s: -1})
@RETURN_NEW(HEAD=TERMS; DATA={cls.one: %(TMP)s, %(RHS)s: -1})
'''
SUB_SYMBOL_VALUE = '@ADD_SYMBOL_VALUE(OBJ=obj; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s; VALUE=-%(VALUE)s)\n'
SUB_NUMBER_SYMBOL = '@SUB_VALUE_SYMBOL(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
SUB_SYMBOL_NUMBER = '@SUB_SYMBOL_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
SUB_SYMBOL_SYMBOL = '''\
if %(LHS)s == %(RHS)s:
    return cls.zero
@RETURN_NEW(HEAD=TERMS; DATA={%(LHS)s: 1, %(RHS)s: -1})
'''
SUB_VALUE_TERMS = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    @NEG_TERMS(OP=%(RHS)s; OPDATA=%(RHSDATA)s)
@NEG_DICT_VALUES(DICT_IN=%(RHSDATA)s; DICT_OUT=pairs)
@ADD_VALUE_DICT(DICT=pairs; VALUE=%(TMP)s)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''
SUB_TERMS_VALUE = '@ADD_VALUE_TERMS(VALUE=-%(VALUE)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
SUB_NUMBER_TERMS = '@SUB_VALUE_TERMS(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
SUB_TERMS_NUMBER = '@SUB_TERMS_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
SUB_TERMS_SYMBOL = '''\
pairs = %(LHSDATA)s.copy()
@ADD_TERM_VALUE_DICT(TERM=%(RHS)s; VALUE=-1; DICT=pairs; DICT_GET=pairs.get; SIGN=+; USIGN=)
@CANONIZE_TERMS_DICT(DICT=pairs)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''
SUB_SYMBOL_TERMS = '''\
@NEG_DICT_VALUES(DICT_IN=%(RHSDATA)s; DICT_OUT=pairs)
@ADD_TERM_VALUE_DICT(TERM=%(LHS)s; VALUE=1; DICT=pairs; DICT_GET=pairs.get; SIGN=+; USIGN=)
@CANONIZE_TERMS_DICT(DICT=pairs)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''
SUB_TERMS_TERMS = '''\
pairs = %(LHSDATA)s.copy()
pairs_get = pairs.get
for t,c in %(RHSDATA)s.iteritems():
    @ADD_TERM_VALUE_DICT(TERM=t; VALUE=-c; DICT=pairs; DICT_GET=pairs_get; SIGN=+; USIGN=)
@CANONIZE_TERMS_DICT(DICT=pairs)
@RETURN_NEW(HEAD=TERMS; DATA=pairs)
'''

#======================================
# MUL macros
#======================================

MUL_ZERO_OP = '''\
if not %(VALUE)s:
    return cls.zero
'''

MUL_VALUE_NUMBER='@RETURN_NEW(HEAD=NUMBER; DATA=%(VALUE)s * %(RHSDATA)s)\n'
MUL_NUMBER_VALUE='@RETURN_NEW(HEAD=NUMBER; DATA=%(LHSDATA)s * %(VALUE)s)\n'
MUL_NUMBER_NUMBER = '@MUL_VALUE_NUMBER(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
MUL_VALUE_SYMBOL = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    return cls.zero
if %(TMP)s==1:
    return %(RHS)s
@RETURN_NEW(HEAD=TERMS; DATA={%(RHS)s: %(TMP)s})
'''
MUL_SYMBOL_VALUE = '@MUL_VALUE_SYMBOL(VALUE=%(VALUE)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
MUL_NUMBER_SYMBOL = '@MUL_VALUE_SYMBOL(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
MUL_SYMBOL_NUMBER = '@MUL_SYMBOL_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
MUL_NUMBER_FACTORS = '@MUL_SYMBOL_VALUE(VALUE=%(LHSDATA)s; LHS=%(RHS)s; LHSDATA=%(RHSDATA)s)\n'
MUL_VALUE_FACTORS = '@MUL_SYMBOL_VALUE(VALUE=%(VALUE)s; LHS=%(RHS)s; LHSDATA=%(RHSDATA)s)\n'
MUL_FACTORS_NUMBER = '@MUL_SYMBOL_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
MUL_FACTORS_VALUE = '@MUL_SYMBOL_VALUE(VALUE=%(VALUE)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
MUL_SYMBOL_SYMBOL = '''\
if %(LHS)s == %(RHS)s:
    pairs = {%(LHS)s: 2}
else:
    pairs = {%(LHS)s: 1, %(RHS)s: 1}
@RETURN_NEW(HEAD=FACTORS; DATA=pairs)
'''
MUL_VALUE_TERMS = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    return cls.zero
pairs = %(RHSDATA)s
if len(pairs)==1:
    t, c = pairs.items()[0]
    c = %(TMP)s * c
    if c==1:
        return t
    @RETURN_NEW(HEAD=TERMS; DATA={t: c})
if %(TMP)s==1:
    return %(RHS)s
pairs = {}
for t,c in %(RHSDATA)s.iteritems():
    pairs[t] = %(TMP)s * c
@NEWINSTANCE(OBJ=obj; HEAD=TERMS; DATA=pairs)
coeff, terms = %(RHS)s._coeff_terms
if terms is not None:
    c = coeff * %(TMP)s
    if not c==1:
        obj._coeff_terms = (c, terms)
else:
    obj._coeff_terms = (%(TMP)s, %(RHS)s)
return obj
'''
MUL_TERMS_VALUE='@MUL_VALUE_TERMS(VALUE=%(VALUE)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
MUL_NUMBER_TERMS = '@MUL_VALUE_TERMS(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
MUL_TERMS_NUMBER = '@MUL_TERMS_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
MUL_TERMS_SYMBOL = '''\
pairs = %(LHSDATA)s
if len(pairs)==1:
    t,c = pairs.items()[0]
    t = t * %(RHS)s
    if t==cls.one:
        return cls(NUMBER, c)
    @RETURN_NEW(HEAD=TERMS; DATA={t: c})
coeff, terms = %(LHS)s._coeff_terms
if terms is not None:
    @NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA={terms:1, %(RHS)s:1})
    @RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s:coeff})
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: 1, %(RHS)s: 1})
'''
MUL_SYMBOL_TERMS = '@MUL_TERMS_SYMBOL(LHS=%(RHS)s; LHSDATA=%(RHSDATA)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
MUL_TERMS_TERMS = '''\
lpairs = %(LHSDATA)s
rpairs = %(RHSDATA)s
if len(lpairs)==1:
    t1,c1 = lpairs.items()[0]
    if len(rpairs)==1:
        t2,c2 = rpairs.items()[0]
        t = t1 * t2
        c = c1 * c2
        if t == cls.one:
            return cls(NUMBER, c)
        if c==1:
            return t
        @RETURN_NEW(HEAD=TERMS; DATA={t: c})
    coeff, terms = %(RHS)s._coeff_terms
    if terms is None:
        return (t1 * %(RHS)s) * c1
    return (t1*terms) * (c1*coeff)
elif len(rpairs)==1:
    t1,c1 = rpairs.items()[0]
    coeff, terms = %(RHS)s._coeff_terms
    if terms is None:
        return (t1 * %(LHS)s) * c1
    return (t1*terms) * (c1*coeff)
lcoeff, lterms = %(LHS)s._coeff_terms
rcoeff, rterms = %(RHS)s._coeff_terms
if lterms is None:
    lterms = %(LHS)s
if rterms is None:
    rterms = %(RHS)s
if lterms==rterms:
    @NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA={lterms: 2})
else:
    @NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA={lterms: 1, rterms: 1})
c = lcoeff * rcoeff
if c==1:
    return %(TMP)s
@RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: c})
'''
MUL_DICT_SYMBOL = '''\
@ADD_TERM_VALUE_DICT(TERM=%(RHS)s; VALUE=1; DICT=%(DICT)s; DICT_GET=%(DICT)s.get; SIGN=+; USIGN=)
@CANONIZE_FACTORS_DICT1(DICT=%(DICT)s)
@RETURN_NEW(HEAD=FACTORS; DATA=%(DICT)s)
'''
MUL_FACTORS_SYMBOL = '''\
pairs = %(LHSDATA)s.copy()
@MUL_DICT_SYMBOL(DICT=pairs; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)
'''
MUL_SYMBOL_FACTORS = '@MUL_FACTORS_SYMBOL(LHS=%(RHS)s; LHSDATA=%(RHSDATA)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
MUL_FACTORS_TERMS = '''\
rpairs = %(RHSDATA)s
if len(rpairs)==1:
    t1,c1 = rpairs.items()[0]
    t = t1 * %(LHS)s
    @RETURN_NEW(HEAD=TERMS; DATA={t: c1})
coeff, terms = %(RHS)s._coeff_terms
if terms is None:
    @MUL_FACTORS_SYMBOL(LHS=%(LHS)s; LHSDATA=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)
pairs = %(LHSDATA)s.copy()
@ADD_TERM_VALUE_DICT(TERM=terms; VALUE=1; DICT=pairs; DICT_GET=pairs.get; SIGN=+; USIGN=)
@CANONIZE_FACTORS_DICT(DICT=pairs; NUMBER=coeff)
@NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA=pairs)
@RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: coeff})
'''
MUL_TERMS_FACTORS = '@MUL_FACTORS_TERMS(LHS=%(RHS)s; LHSDATA=%(RHSDATA)s; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
MUL_FACTORS_FACTORS = '''\
pairs = %(LHSDATA)s.copy()
pairs_get = pairs.get
number = 1
for t,c in %(RHSDATA)s.iteritems():
    @MUL_FACTOR_VALUE_DICT(FACTOR=t; SIGN=+; USIGN=; VALUE=c; DICT=pairs; DICT_GET=pairs_get; NUMBER=number)
@CANONIZE_FACTORS_DICT(DICT=pairs; NUMBER=number)
if number == 1:
    @RETURN_NEW(HEAD=FACTORS; DATA=pairs)
@NEWINSTANCE(OBJ=obj; HEAD=FACTORS; DATA=pairs)
@RETURN_NEW(HEAD=TERMS; DATA={obj: number})
'''

#======================================
# DIV macros
#======================================

DIV_VALUE_NUMBER='@RETURN_NEW2(HEAD=NUMBER; DATA=div(%(VALUE)s, %(RHSDATA)s, cls))\n'
DIV_NUMBER_VALUE='@RETURN_NEW2(HEAD=NUMBER; DATA=div(%(LHSDATA)s, %(VALUE)s, cls))\n'
DIV_NUMBER_NUMBER = '@DIV_VALUE_NUMBER(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
DIV_VALUE_SYMBOL = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    return cls.zero
@NEWINSTANCE(OBJ=obj2; HEAD=FACTORS; DATA={%(RHS)s: -1})
if %(TMP)s==1:
    return obj2
@RETURN_NEW(HEAD=TERMS; DATA={obj2: %(TMP)s})
'''
DIV_SYMBOL_VALUE = '@MUL_VALUE_SYMBOL(VALUE=div(1, %(VALUE)s, cls); RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)\n'
DIV_NUMBER_SYMBOL = '@DIV_VALUE_SYMBOL(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
DIV_SYMBOL_NUMBER = '@DIV_SYMBOL_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
DIV_TERMS_VALUE = '@MUL_TERMS_VALUE(LHS=%(LHS)s; LHSDATA=%(LHSDATA)s; VALUE=div(1,%(VALUE)s,cls))\n'

DIV_VALUE_TERMS = '''\
%(TMP)s = %(VALUE)s
if not %(TMP)s:
    return cls.zero
pairs = %(RHSDATA)s
if len(pairs)==1:
    t, c = pairs.items()[0]
    c = div(%(VALUE)s, c, cls)
    t = 1/t
    if c==1:
        return t
    if t==cls.one:
        return cls(NUMBER, c)
    @RETURN_NEW(HEAD=TERMS; DATA={t: c})
@NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA={%(RHS)s: -1})
if %(VALUE)s==1:
    return %(TMP)s
@RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: %(VALUE)s})
'''

DIV_FACTORS_VALUE = '@MUL_FACTORS_VALUE(LHS=%(LHS)s; LHSDATA=%(LHSDATA)s; VALUE=div(1,%(VALUE)s,cls))\n'
DIV_VALUE_FACTORS = '''
pairs = %(RHSDATA)s
if len(pairs)==1:
    t, c = pairs.items()[0]
    c = -c
    if c==1:
        return t * %(VALUE)s
    new_pairs = {t: c}
else:
    @NEG_DICT_VALUES(DICT_IN=pairs; DICT_OUT=new_pairs)
@NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA=new_pairs)
if %(VALUE)s==1:
    return %(TMP)s
@RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: %(VALUE)s})
'''
DIV_NUMBER_TERMS = '@DIV_VALUE_TERMS(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
DIV_TERMS_NUMBER = '@DIV_TERMS_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
DIV_NUMBER_FACTORS = '@DIV_VALUE_FACTORS(VALUE=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)\n'
DIV_FACTORS_NUMBER = '@DIV_FACTORS_VALUE(VALUE=%(RHSDATA)s; LHS=%(LHS)s; LHSDATA=%(LHSDATA)s)\n'
DIV_SYMBOL_SYMBOL = '''\
if %(LHS)s == %(RHS)s:
    return cls.one
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: 1, %(RHS)s: -1})
'''
DIV_TERMS_SYMBOL = '''
pairs = %(LHSDATA)s
if len(pairs)==1:
    t, c = pairs.items()[0]
    if t==%(RHS)s:
        return cls(NUMBER, c)
    if t.head is FACTORS:
        %(TMP)s = t / %(RHS)s
        @RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: c})
    @NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA={t:1, %(RHS)s: -1})
    @RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: c})
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: 1, %(RHS)s: -1})
'''
DIV_TERMS_TERMS = '''
if %(LHS)s==%(RHS)s:
    return cls.one
lpairs = %(LHSDATA)s
rpairs = %(RHSDATA)s
if len(lpairs)==1:
    t1, c1 = lpairs.items()[0]
    if len(rpairs)==1:
        t2, c2 = rpairs.items()[0]
        c = div(c1, c2, cls)
        if t2==t1:
            return cls(NUMBER, c)
        if c==1:
            return t1 / t2
            #@RETURN_NEW(HEAD=FACTORS; DATA={t1:1, t2:-1})
        return (t1 / t2) * c
        #@NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA={t1:1, t2:-1})
    else:
        %(TMP)s = t1 / %(RHS)s
    @RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s:c1})
elif len(rpairs)==1:
    t2, c2 = rpairs.items()[0]
    c = div(1, c2, cls)
    if t2==%(LHS)s:
        return cls(NUMBER, c)
    %(TMP)s = %(LHS)s / t2
    @RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s:c})
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s:1, %(RHS)s:-1})
'''
DIV_SYMBOL_TERMS = '''\
pairs = %(RHSDATA)s
if len(pairs)==1:
    t,c = pairs.items()[0]
    if %(LHS)s==t:
        return cls(NUMBER, div(1, c, cls))
    @NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA={%(LHS)s:1, t:-1})
    @RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: div(1, c, cls)})
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s:1, %(RHS)s:-1})
'''
DIV_SYMBOL_FACTORS = '''\
pairs = %(RHSDATA)s
if len(pairs)==1:
    t, c = pairs.items()[0]
    if t==%(LHS)s:
        c = 1 - c
        if not c:
            return cls.one
        if c==1:
            return t
        else:
            @RETURN_NEW(HEAD=FACTORS; DATA={t: c})
    @RETURN_NEW(HEAD=FACTORS; DATA={t: -c, %(LHS)s: 1})
@NEG_DICT_VALUES(DICT_IN=%(RHSDATA)s; DICT_OUT=pairs)
@MUL_DICT_SYMBOL(DICT=pairs; RHS=%(LHS)s; RHSDATA=%(LHSDATA)s)
'''
DIV_TERMS_FACTORS = '''\
lpairs = %(LHSDATA)s
if len(lpairs)==1:
    t, c = lpairs.items()[0]
    t = t / %(RHS)s
    if t==cls.one:
        return cls(NUMBER, c)
    head, data = t.pair
    if head is NUMBER:
        @RETURN_NEW(HEAD=NUMBER; DATA=data * c)
    elif head is TERMS:
        @MUL_TERMS_VALUE(LHS=t; LHSDATA=data; VALUE=c)
    else:
        @MUL_SYMBOL_VALUE(LHS=t; LHSDATA=data; VALUE=c)
@DIV_SYMBOL_FACTORS(LHS=%(LHS)s; LHSDATA=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)
'''
DIV_DICT_SYMBOL = '''\
@ADD_TERM_VALUE_DICT(TERM=%(RHS)s; VALUE=-1; DICT=%(DICT)s; DICT_GET=%(DICT)s.get; SIGN=+; USIGN=)
@CANONIZE_FACTORS_DICT1(DICT=%(DICT)s)
@RETURN_NEW(HEAD=FACTORS; DATA=%(DICT)s)
'''
DIV_FACTORS_SYMBOL = '''\
pairs = %(LHSDATA)s.copy()
@DIV_DICT_SYMBOL(RHS=%(RHS)s; RHSDATA=%(RHSDATA)s; DICT=pairs)
'''
DIV_FACTORS_TERMS = '''\
rpairs = %(RHSDATA)s
if len(rpairs)==1:
    t, c = rpairs.items()[0]
    t = %(LHS)s / t
    c = div(1, c, cls)
    if t==cls.one:
        return cls(NUMBER, c)
    head, data = t.pair
    if head is NUMBER:
        @RETURN_NEW(HEAD=NUMBER; DATA=data * c)
    elif head is TERMS:
        @MUL_TERMS_VALUE(LHS=t; LHSDATA=data; VALUE=c)
    else:
        @MUL_SYMBOL_VALUE(LHS=t; LHSDATA=data; VALUE=c)
@DIV_FACTORS_SYMBOL(LHS=%(LHS)s; LHSDATA=%(LHSDATA)s; RHS=%(RHS)s; RHSDATA=%(RHSDATA)s)
'''
DIV_FACTORS_FACTORS = '''\
pairs = %(LHSDATA)s.copy()
pairs_get = pairs.get
number = 1
for t,c in %(RHSDATA)s.iteritems():
    @MUL_FACTOR_VALUE_DICT(FACTOR=t; SIGN=-; USIGN=-; VALUE=c; DICT=pairs; DICT_GET=pairs_get; NUMBER=number)
@CANONIZE_FACTORS_DICT(DICT=pairs; NUMBER=number)
if number==1:
    @RETURN_NEW(HEAD=FACTORS; DATA=pairs)
@NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA=pairs)
@RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: number})
'''

#======================================
# POW macros
#======================================

POW_NUMBER_INT = '''\
if %(VALUE)s < 0:
    @RETURN_NEW(HEAD=NUMBER; DATA=div(1, (%(LHSDATA)s)**(-%(VALUE)s), cls))
@RETURN_NEW(HEAD=NUMBER; DATA=(%(LHSDATA)s)**(%(VALUE)s))
'''
POW_TERMS_INT = '''\
pairs = %(LHSDATA)s
if len(pairs)==1:
    t,c = pairs.items()[0]
    if %(VALUE)s < 0:
        c = div(1, c**(-%(VALUE)s), cls)
    else:
        c = c ** (%(VALUE)s)
    t = t**(%(VALUE)s)
    if c==1:
        return t
    @RETURN_NEW(HEAD=TERMS; DATA={t:c})
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: %(VALUE)s})
'''
POW_FACTORS_INT = '''\
@MUL_DICT_VALUES(DICT_IN=%(LHSDATA)s; DICT_OUT=pairs; OP=%(VALUE)s)
if len(pairs)==1:
    t, c = pairs.items()[0]
    if c==1:
        return t
@RETURN_NEW(HEAD=FACTORS; DATA=pairs)
'''
POW_SYMBOL_INT = '''\
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: %(VALUE)s})
'''

POW_NUMBER_VALUE = '''\
z, sym = try_power(%(LHSDATA)s, %(VALUE)s)
if not sym:
    @RETURN_NEW(HEAD=NUMBER; DATA=z)
factors = {}
for t,c in sym:
    factors[cls(NUMBER, t)] = c
@NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA=factors)
if z==1:
    return %(TMP)s
@RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: z})
'''

POW_TERMS_FRAC = '''\
pairs = %(LHSDATA)s
if len(pairs)==1:
    t, c = pairs.items()[0]
    if isinstance(c, numbertypes) and not c==-1:
        if c < 0:
            z, sym = try_power(-c, %(VALUE)s)
            factors = {-t: %(VALUE)s}
        else:
            z, sym = try_power(c, %(VALUE)s)
            factors = {t: %(VALUE)s}
        for t,c in sym:
            factors[cls(NUMBER, t)] = c
        @NEWINSTANCE(OBJ=%(TMP)s; HEAD=FACTORS; DATA=factors)
        if z==1:
            return %(TMP)s
        @RETURN_NEW(HEAD=TERMS; DATA={%(TMP)s: z})
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: %(VALUE)s})
'''

POW_SYMBOL_FRAC = '''\
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: %(VALUE)s})
'''

POW_FACTORS_SYMBOL = '''\
pairs = %(LHSDATA)s
if len(pairs)==1:
    t, c = pairs.items()[0]
    if type(c) in inttypes_set:
        @RETURN_NEW(HEAD=FACTORS; DATA={t: %(RHS)s * c})
@RETURN_NEW(HEAD=FACTORS; DATA={%(LHS)s: %(RHS)s})
'''

def generate_if_blocks(heads, prefix='', tab=' '*4):
    lines = []
    lapp = lines.append
    for h1 in heads:
        if h1==heads[-1]:
            lapp('else:')
        elif h1==heads[0]:
            lapp('if lhead is %s:' % (h1))
        else:
            lapp('elif lhead is %s:' % (h1))
        for h2 in heads:
            if h2==heads[-1]:
                lapp(tab+'else:')
            elif h2==heads[0]:
                lapp(tab+'if rhead is %s:' % (h2))
            else:
                lapp(tab+'elif rhead is %s:' % (h2))
            lapp(tab*2 + '@%%(OP)s_%s_%s(LHS=self; LHSDATA=ldata; RHS=other; RHSDATA=rdata)' % (h1, h2))
    return prefix + ('\n'+prefix).join(lines)

OP3_TEMPLATE = '''
def %(op)s_method(self, other, NUMBER=NUMBER, TERMS=TERMS, FACTORS=FACTORS):
    cls = type(self)
    lhead, ldata = self.pair
    if type(other) is not cls:
        if isinstance(other, cls.coefftypes):
            if lhead is NUMBER:
                @%(OP)s_NUMBER_VALUE(VALUE=other; LHS=self; LHSDATA=ldata)
            elif lhead is TERMS:
                @%(OP)s_TERMS_VALUE(VALUE=other; LHS=self; LHSDATA=ldata)
            else:
                @%(OP)s_SYMBOL_VALUE(VALUE=other; LHS=self; LHSDATA=ldata)
        other = cls.convert(other, False)
        if other is NotImplemented:
            return other
    rhead, rdata = other.pair
''' + generate_if_blocks(['NUMBER', 'TERMS', 'SYMBOL'], prefix=' '*4)

OP4_TEMPLATE = '''
def %(op)s_method(self, other, NUMBER=NUMBER, TERMS=TERMS, FACTORS=FACTORS):
    cls = type(self)
    lhead, ldata = self.pair
    if type(other) is not cls:
        if isinstance(other, cls.coefftypes):
            if lhead is NUMBER:
                @%(OP)s_NUMBER_VALUE(VALUE=other; LHS=self; LHSDATA=ldata)
            elif lhead is TERMS:
                @%(OP)s_TERMS_VALUE(VALUE=other; LHS=self; LHSDATA=ldata)
            elif lhead is FACTORS:
                @%(OP)s_FACTORS_VALUE(VALUE=other; LHS=self; LHSDATA=ldata)
            else:
                @%(OP)s_SYMBOL_VALUE(VALUE=other; LHS=self; LHSDATA=ldata)
        other = cls.convert(other, False)
        if other is NotImplemented:
            return other
    rhead, rdata = other.pair
''' + generate_if_blocks(['NUMBER', 'TERMS', 'FACTORS', 'SYMBOL'], prefix=' '*4)

def main():
    f = open(targetfile_py, 'w')
    print >> f, template
    print >> f, preprocess('''

def expand_mul_method(cls, self, other):
    lhead, ldata = self.pair
    rhead, rdata = other.pair
    if lhead is FACTORS:
        if rhead is FACTORS:
            @MUL_FACTORS_FACTORS(LHS=self; LHSDATA=ldata; RHS=other; RHSDATA=rdata)
        elif rhead is NUMBER:
            return self # other must be one
        else:
            @MUL_FACTORS_SYMBOL(LHS=self; LHSDATA=ldata; RHS=other; RHSDATA=rdata)
    elif lhead is NUMBER:
        return other # self must be one
    else:
        if rhead is FACTORS:
            @MUL_SYMBOL_FACTORS(LHS=self; LHSDATA=ldata; RHS=other; RHSDATA=rdata)
        elif rhead is NUMBER:
            return self # other must be one
        else:
            @MUL_SYMBOL_SYMBOL(LHS=self; LHSDATA=ldata; RHS=other; RHSDATA=rdata)

#def rsub_method(self, other, NUMBER=NUMBER, TERMS=TERMS, FACTORS=FACTORS):
#    cls = type(self)
#    lhead, ldata = self.pair
#    if isinstance(other, cls.coefftypes):
#        if lhead is NUMBER:
#            @SUB_VALUE_NUMBER(VALUE=other; RHS=self; RHSDATA=ldata)
#        elif lhead is TERMS:
#            @SUB_VALUE_TERMS(VALUE=other; RHS=self; RHSDATA=ldata)
#        else:
#            @SUB_VALUE_SYMBOL(VALUE=other; RHS=self; RHSDATA=ldata)
#    other = cls.convert(other, False)
#    if other is NotImplemented:
#        return other
#    return other - self

def rdiv_method(self, other, NUMBER=NUMBER, TERMS=TERMS, FACTORS=FACTORS):
    cls = type(self)
    lhead, ldata = self.pair
    if isinstance(other, cls.coefftypes):
        if lhead is NUMBER:
            @DIV_VALUE_NUMBER(VALUE=other; RHS=self; RHSDATA=ldata)
        elif lhead is TERMS:
            @DIV_VALUE_TERMS(VALUE=other; RHS=self; RHSDATA=ldata)
        elif lhead is FACTORS:
            @DIV_VALUE_FACTORS(VALUE=other; RHS=self; RHSDATA=ldata)
        else:
            @DIV_VALUE_SYMBOL(VALUE=other; RHS=self; RHSDATA=ldata)
    other = cls.convert(other, False)
    if other is NotImplemented:
        return other
    return other / self

''', globals())

    #print >> f, preprocess(OP3_TEMPLATE % (dict(op='add', OP='ADD')), globals())
    #print >> f, preprocess(OP3_TEMPLATE % (dict(op='sub', OP='SUB')), globals())
    #print >> f, preprocess(OP4_TEMPLATE % (dict(op='mul', OP='MUL')), globals())
    print >> f, preprocess(OP4_TEMPLATE % (dict(op='div', OP='DIV')), globals())

    f.close()


if __name__=='__main__':
    main()



from __future__ import with_statement

from sympycore.calculus import Calculus as A
from sympycore.calculus.algebra import I
from sympycore.calculus.infinity import oo, zoo, undefined
from sympycore.calculus import Number, Symbol
from sympycore.calculus.functions.elementary import Sin, Cos, Tan, Cot, pi, E, Exp, Log
from sympycore.calculus.relational import Assumptions
from sympycore import FD, D

def test_exp_log():
    assert Exp(1) == E
    assert Exp(0) == 1
    assert Log(0) == -oo
    assert Log(1) == 0
    assert Log(oo) == oo
    assert Log(2,3) == Log(2)/Log(3)
    assert Log(0,3) == (-oo)/Log(3)
    assert Log(100,10) == 2
    assert Log(65536,2) == 16
    assert Log(101,10) != 2
    assert Log(100,0) == 0
    assert Log(5,pi) == Log(5)/Log(pi)
    assert Log(-3) == I*pi + Log(3)
    assert Log(-pi) == I*pi + Log(pi)
    assert Log(I) == I*pi/2
    assert Log(-I) == -I*pi/2
    assert Log(3*I) == Log(3) + I*pi/2
    assert Log(-3*I) == Log(3) - I*pi/2
    assert str(Log(1+I)) == 'Log(1 + I)'
    assert Log(5**Number(1,3)) == Number(1,3)*Log(5), str(Log(5**Number(1,3)))
    assert Log(5**(3+2*I)) != (3+2*I)*Log(5)

def test_log_assumptions():
    x = Symbol('x')
    assert Log(x**2) != 2*Log(x)
    with Assumptions([x > 0]):
        assert Log(x**2) == 2*Log(x),str(Log(x**2))

def test_trig_values():
    sqrt2 = A.convert('2**(1/2)')
    sqrt3 = A.convert('3**(1/2)')
    assert Sin(0) == 0
    assert Sin(pi) == 0
    assert Sin(4*pi) == 0
    assert Sin(3*pi/2) == -1, `Sin(3*pi/2)`
    assert Sin(5*pi/2) == 1
    assert Sin(pi/3) == sqrt3/2, `Sin(pi/3).pair, (sqrt3/2).pair`
    assert Sin(pi/2) == 1
    assert Cos(0) == 1
    assert Cos(pi) == -1
    assert Cos(8*pi) == 1
    assert Cos(-9*pi) == -1
    assert Cos(pi/2) == 0
    assert Cos(3*pi/2) == 0
    assert Cos(11*pi/2) == 0
    assert Cos(pi/12) == (1 + sqrt3) / (2 * sqrt2)
    assert Tan(7*pi/12) == Sin(7*pi/12)/Cos(7*pi/12)
    assert Tan(pi/2) == zoo,`Tan(pi/2)`
    assert Tan(pi) == 0
    assert Cot(pi/2) == 0
    assert Cot(pi) == zoo
    assert str(Sin(oo)) == 'Sin(oo)'
    assert Sin(undefined) == undefined

def test_trig_symmetry():
    x = Symbol('x')
    assert Sin(-x) == -Sin(x),`Sin(-x).pair, (-Sin(x)).pair`
    assert Cos(-x) == Cos(x)
    assert Tan(-x) == -Tan(x)
    assert Cot(-x) == -Cot(x)
    assert Sin(x+pi) == -Sin(x),`Sin(x+pi)`
    assert Sin(x+2*pi) == Sin(x)
    assert Sin(x+3*pi) == -Sin(x)
    assert Sin(x+4*pi) == Sin(x)
    assert Sin(x-5*pi) == -Sin(x)
    assert Cos(x+pi) == -Cos(x)
    assert Cos(x+2*pi) == Cos(x)
    assert Cos(x+3*pi) == -Cos(x)
    assert Cos(x+4*pi) == Cos(x)
    assert Cos(x-5*pi) == -Cos(x)
    assert Tan(x+pi) == Tan(x)
    assert Tan(x-3*pi) == Tan(x)
    assert Cot(x+pi) == Cot(x)
    assert Cot(x-3*pi) == Cot(x)
    assert Sin(pi/2-x) == Cos(x)
    assert Sin(3*pi/2-x) == -Cos(x)
    assert Sin(5*pi/2-x) == Cos(x)
    assert Cos(pi/2-x) == Sin(x)
    assert Cos(3*pi/2-x) == -Sin(x)
    assert Cos(5*pi/2-x) == Sin(x)
    assert Tan(pi/2-x) == Cot(x)
    assert Tan(3*pi/2-x) == Cot(x)
    assert Tan(5*pi/2-x) == Cot(x)
    assert Cot(pi/2-x) == Tan(x)
    assert Cot(3*pi/2-x) == Tan(x)
    assert Cot(5*pi/2-x) == Tan(x)
    assert Sin(pi/2+x) == Cos(x)
    assert Cos(pi/2+x) == -Sin(x)
    assert Tan(pi/2+x) == -Cot(x)
    assert Cot(pi/2+x) == -Tan(x)

def test_trig_diff():
    x = Symbol('x')
    assert Sin(x).diff(x) == Cos(x), `Sin(x).diff(x)`
    assert Cos(x).diff(x) == -Sin(x)
    assert Sin(2*x).diff(x) == 2*Cos(2*x)

    assert Tan(x).diff(x) == 1+Tan(x)**2
    assert Cot(x).diff(x) == -1-Cot(x)**2

    assert Log(x).diff(x) == 1/x
    assert Exp(x).diff(x) == Exp(x)


    assert (x*Sin(x)).diff(x) == x*Cos(x) + Sin(x)

def test_trig_diff_intorder():
    x = Symbol('x')
    assert Sin(x).diff(x,2) == -Sin(x)
    assert Sin(x).diff(x,3) == -Cos(x)
    assert Sin(x).diff(x,4) == Sin(x)
    assert Sin(x).diff(x,4000000) == Sin(x)

    assert Cos(x).diff(x,2) == -Cos(x)
    assert Cos(x).diff(x,3) == Sin(x)
    assert Cos(x).diff(x,4) == Cos(x)
    assert Cos(x).diff(x,4000000) == Cos(x)

    assert Tan(x).diff(x,2) == Tan(x).diff(x).diff(x)
    assert Cot(x).diff(x,2) == Cot(x).diff(x).diff(x)

def test_trig_diff_symorder():
    x = Symbol('x')
    n = Symbol('n')
    assert Sin(x).diff(x,n) == Sin(x + pi*n/2),`Sin(x).diff(x,n)`
    assert Cos(x).diff(x,n) == Cos(x + pi*n/2),`Cos(x).diff(x,n)`
    assert Tan(x).diff(x,n) == (FD[0]**n)(Tan)(x),`Tan(x).diff(x,n)`
    assert Cot(x).diff(x,n) == (FD[0]**n)(Cot)(x),`Cot(x).diff(x,n)`



import pymaclab as pm
import pymaclab.modfiles.models as models
import pymaclab.modfiles.templates.wheezy_template as template
import copy


def test_cee():

    # Instantiate model in the normal way
    cee = pm.newMOD(models.testing.cee,mesg=True)
    params = copy.deepcopy(cee.template_paramdic)
    
    # Now use the template to do it again
    modstr = template.render(params)
    cee2 = pm.newMOD(modstr,mesg=True)
    
    # Check if the two template_paramdics are identical
    for keyo in cee.template_paramdic.keys():
        print "Now testing key: ",keyo
        if keyo == 'sigma':
            assert cee.template_paramdic[keyo].all() == cee2.template_paramdic[keyo].all()
        elif keyo == 'paramdic':
            for keyo2 in cee.template_paramdic[keyo].keys():
                assert round(cee.template_paramdic[keyo][keyo2],6) == round(cee2.template_paramdic[keyo][keyo2],6)
        else:
            assert cee.template_paramdic[keyo] == cee2.template_paramdic[keyo]


def test_others():

    # Do for a couple of other models as well
    modelli = []
    for elem in dir(models.stable):
        if '__' not in elem: modelli.append(elem)
    # Remove models which need external file or otherwise don't work
    modelli.remove('rbc1_ext')
    modelli.remove('rbc1_extss')
    modelli.remove('prog')
    modelli.remove('jermann98_ext')
    
    for modelo in modelli:
        print "Now testing for model: ",modelo
        exec(modelo+" = pm.newMOD(models.stable."+modelo+",mesg=True)")
        params = copy.deepcopy(eval(modelo+".template_paramdic"))
        modstr = template.render(params)
        exec(modelo+'_alt'+" = pm.newMOD(modstr,mesg=True)")
        # Check if the two template_paramdics are identical
        for keyo in eval(modelo+".template_paramdic.keys()"):
            print "Now testing key: ",keyo
            if keyo == 'sigma':
                assert eval(modelo+".template_paramdic[keyo].all()") == eval(modelo+"_alt"+".template_paramdic[keyo].all()")
            elif keyo == 'paramdic':
                for keyo2 in eval(modelo+".template_paramdic[keyo].keys()"):
                    assert round(eval(modelo+".template_paramdic[keyo][keyo2]"),6) == round(eval(modelo+"_alt"+".template_paramdic[keyo][keyo2]"),6)
            elif keyo == 'ssidic' and eval(modelo+".template_paramdic[keyo]") != False:
                for keyo2 in eval(modelo+".template_paramdic[keyo].keys()"):
                    assert round(eval(modelo+".template_paramdic[keyo][keyo2]"),6) == round(eval(modelo+"_alt"+".template_paramdic[keyo][keyo2]"),6)
            else:
                assert eval(modelo+".template_paramdic[keyo]") == eval(modelo+"_alt"+".template_paramdic[keyo]")    


     




__all__ = ['NUMBER']

import re

from .base import Head, Expr, heads_precedence, AtomicHead, Pair

from ..core import init_module
init_module.import_heads()
init_module.import_numbers()
init_module.import_lowlevel_operations()

_is_number = re.compile(r'\A[-]?\d+\Z').match
_is_neg_number = re.compile(r'\A-\d+([/]\d+)?\Z').match
_is_rational = re.compile(r'\A\d+[/]\d+\Z').match

class NumberHead(AtomicHead):
    """
    NumberHead is a head for symbols, data can be any Python object.
    """

    def __repr__(self): return 'NUMBER'

    def new(self, cls, data):
        if isinstance(data, Infinity):
            return data
        return cls(NUMBER, data)

    def reevaluate(self, cls, data):
        return cls(self, data)

    def is_data_ok(self, cls, data):
        return

    def nonzero(self, cls, data):
        return data

    def to_SPARSE_POLY(self, cls, data, expr):
        """
        Obsolete method:  SPARSE_POLY will be removed in future.
        """
        return cls(data)

    def to_TERM_COEFF_DICT(self, cls, data, expr):
        return expr

    def to_ADD(self, cls, data, expr):
        return expr
        
    def to_EXP_COEFF_DICT(self, cls, data, expr, variables = None):
        if variables is None:
            variables = ()
        if data:
            return cls(EXP_COEFF_DICT, Pair(variables, {EXP_COEFF_DICT.make_exponent(0, variables):data}))
        return cls(EXP_COEFF_DICT, Pair(variables, {}))

    def data_to_str_and_precedence(self, cls, data):
        if isinstance(data, complextypes):
            r, i = data.real, data.imag
            if r!=0 and i!=0:
                return str(data), heads_precedence.ADD
            if r==0:
                if i<0:
                    return str(data), heads_precedence.NEG
                return str(data), heads_precedence.NUMBER
            return self.data_to_str_and_precedence(self, cls, r)
        elif isinstance(data, rationaltypes):
            if data < 0:
                return str(data), heads_precedence.NEG
            return str(data), heads_precedence.DIV
        elif isinstance(data, realtypes):
            if data < 0:
                return str(data), heads_precedence.NEG
            return str(data), heads_precedence.NUMBER
        elif isinstance(data, Expr):
            h, d = data.pair
            return h.data_to_str_and_precedence(type(data), d)
        return str(data), 0.0 # force parenthesis

    def non_commutative_mul(self, cls, lhs, rhs):
        head, data = rhs.pair
        if head is NUMBER:
            return cls(NUMBER, lhs.data * data)
        # Numbers are ring commutants:
        return rhs.head.non_commutative_mul(cls, rhs, lhs)

    def commutative_mul_number(self, cls, lhs, rhs):
        return cls(NUMBER, lhs.data * rhs)

    non_commutative_rmul_number = commutative_mul_number

    def commutative_div_number(self, cls, lhs, rhs):
        r = number_div(cls, lhs.data, rhs)
        if rhs==0:
            return r * lhs
        return cls(NUMBER, number_div(cls, lhs.data, rhs))

    def commutative_rdiv_number(self, cls, lhs, rhs):
        r = number_div(cls, rhs, lhs.data)
        if rhs==0:
            return r * lhs
        return cls(NUMBER, r)

    def commutative_div(self, cls, lhs, rhs):
        return rhs.head.commutative_rdiv_number(cls, rhs, lhs.data)

    non_commutative_mul_number = commutative_mul_number

    def commutative_mul(self, cls, lhs, rhs):
        head, data = rhs.pair
        if head is NUMBER:
            return cls(NUMBER, lhs.data * data)
        return rhs.head.commutative_mul(cls, rhs, lhs)

    inplace_commutative_mul = commutative_mul

    def term_coeff(self, cls, expr):
        if isinstance(expr, Expr):
            return cls(NUMBER, 1), expr.data
        return cls(NUMBER, 1), expr

    def neg(self, cls, expr):
        return cls(self, -expr.data)

    def add(self, cls, lhs, rhs):
        ldata = lhs.data
        if ldata==0:
            return rhs
        h, d = rhs.pair
        if h is NUMBER:
            return cls(NUMBER, ldata + d)
        elif h is SYMBOL or h is APPLY or h is CALLABLE:
            return cls(TERM_COEFF_DICT, {cls(NUMBER,1): ldata, rhs:1})
        elif h is ADD:
            terms = []
            for term in d:
                h1, c = term.pair
                if h1 is NUMBER:
                    c = lhs.data + c
                    if c:
                        terms.append(cls(NUMBER, c))
                else:
                    terms.append(term)
            if not terms:
                return cls(NUMBER, 0)
            if len(terms)==1:
                return terms[0]
            return cls(ADD, terms)
        elif h is TERM_COEFF:
            term, coeff = d
            return cls(TERM_COEFF_DICT, {term:coeff, cls(NUMBER,1):lhs.data})
        elif h is POW or h is BASE_EXP_DICT:
            return cls(TERM_COEFF_DICT, {cls(NUMBER,1):lhs.data, rhs:1})
        elif h is TERM_COEFF_DICT:
            data = d.copy()
            dict_add_item(cls, data, cls(NUMBER,1), lhs.data)
            return term_coeff_dict_new(cls, data)
        elif h is MUL:
            return cls(ADD, [lhs, rhs])
        raise NotImplementedError(`self, lhs.pair, rhs.pair`)

    inplace_add = add

    def add_number(self, cls, lhs, rhs):
        return cls(NUMBER, lhs.data + rhs) if rhs else lhs

    def sub_number(self, cls, lhs, rhs):
        return cls(NUMBER, lhs.data - rhs) if rhs else lhs

    def sub(self, cls, lhs, rhs):
        return lhs + (-rhs)

    def pow(self, cls, base, exp):
        h, d = exp.pair
        if h is NUMBER and type(d) in numbertypes_set:
            return self.pow_number(cls, base, d)
        return pow_new(cls, (base, exp))

    def pow_number(self, cls, base, exp):
        if exp==1:
            return base
        if exp==0:
            return cls(NUMBER, 1)
        bd = base.data
        if bd==1:
            return base
        if bd==0:
            if exp>0: return base
            # exp is negative
            return number_div(cls, 1, 0)
        r, l = try_power(bd, exp)
        if not l:
            return cls(NUMBER, r)
        if len(l)==1:
            b, e = l[0]
            if r==1:
                return cls(POW, (cls(NUMBER, b), e))
            return cls(TERM_COEFF, (cls(POW, (cls(NUMBER, b), e)), r))
        d = {}
        for b, e in l:
            d[cls(NUMBER, b)] = e
        if r == 1:
            return cls(BASE_EXP_DICT, d)
        return cls(TERM_COEFF, (cls(BASE_EXP_DICT, d), r))

    def expand(self, cls, expr):
        return expr

    def diff(self, cls, data, expr, symbol, order, cache={}):
        #if isinstance(data, Expr):
        #    return cls(NUMBER, data.head.diff(type(data), data.data, data, symbol, order))
        return cls(NUMBER, 0)

    def apply(self, cls, data, func, args):
        if isinstance(data, cls):
            return data
        return cls(NUMBER, data)

    def integrate_indefinite(self, cls, data, expr, x):
        if data==0:
            return expr
        if data==1:
            return cls(SYMBOL, x)
        return cls(TERM_COEFF, (cls(SYMBOL, x), data))

    def integrate_definite(self, cls, data, expr, x, a, b):
        if data==0:
            return expr
        if data==1:
            return b-a
        return (b-a)*data

    def algebra_pos(self, Algebra, expr):
        return expr

    def algebra_neg(self, Algebra, expr):
        if Algebra.algebra_options.get('evaluate_addition'):
            return Algebra(NUMBER, -expr.data)
        return Algebra(NEG, expr)

    def algebra_add_number(self, Algebra, lhs, rhs, inplace):
        if Algebra.algebra_options.get('evaluate_addition'):
            return Algebra(NUMBER, lhs.data + rhs)
        return Algebra(ADD, [lhs, Algebra(NUMBER, rhs)])

    def algebra_add(self, Algebra, lhs, rhs, inplace):
        ldata = lhs.data
        rhead, rdata = rhs.pair
        if Algebra.algebra_options.get('is_additive_group_commutative'):
            if Algebra.algebra_options.get('evaluate_addition'):
                if rhead is ADD or rhead is EXP_COEFF_DICT or rhead is MUL or rhead is NEG:
                    rhs = rhead.to_TERM_COEFF_DICT(Algebra, rdata, rhs)
                    rhead, rdata = rhs.pair
                if not ldata:
                    return rhs
                if rhead is NUMBER:
                    return Algebra(NUMBER, lhs.data + rdata)
                if rhead is SYMBOL:
                    return Algebra(TERM_COEFF_DICT, {Algebra(NUMBER, 1): ldata, rhs:1})
                if rhead is TERM_COEFF or rhead is TERM_COEFF_DICT:
                    return rhead.algebra_add_number(Algebra, rhs, ldata, inplace)
                return super(type(self), self).algebra_add(Algebra, lhs, rhs, inplace)
            return Algebra(ADD, [lhs, rhs])
        else:
            if rhead is ADD:
                data = [lhs] + rdata
            elif rhead is TERM_COEFF_DICT or rhead is EXP_COEFF_DICT:
                data = [lhs] + rhs.to(ADD).data
            else:
                data = [lhs, rhs]
            if Algebra.algebra_options.get('evaluate_addition'):
                ADD.combine_add_list(Algebra, data)
            return add_new(Algebra, data)

    def algebra_mul_number(self, Algebra, lhs, rhs, inplace):
        if Algebra.algebra_options.get('evaluate_addition'):
            return Algebra(NUMBER, lhs.data * rhs)
        return mul_new(Algebra, [lhs, Algebra(NUMBER, rhs)])
    
    def algebra_mul(self, Algebra, lhs, rhs, inplace):
        if Algebra.algebra_options.get('evaluate_addition'):
            return rhs.head.algebra_mul_number(Algebra, rhs, lhs.data, inplace)
        return mul_new(Algebra, [lhs, rhs])

    def algebra_pow_number(self, Algebra, lhs, rhs, inplace):
        if Algebra.algebra_options.get('evaluate_multiplication'):
            if rhs==1:
                return lhs
            if not rhs:
                return Algebra(NUMBER, 1)
            ldata = lhs.data
            if ldata==1:
                return lhs
            if not ldata:
                if rhs>0:
                    return lhs
                return number_div(cls, 1, 0)
            r, l = try_power(ldata, rhs)
            if not l:
                return Algebra(NUMBER, r)
            if Algebra.algebra_options.get('is_multiplicative_group_commutative'):
                d = {}
                for b, e in l:
                    b = Algebra(NUMBER, b)
                    d[b] = e
                p = base_exp_dict_new(Algebra, d)
                return term_coeff_new(Algrbra, (p, r))
            else:
                d = []
                for b, e in l:
                    b = Algebra(NUMBER, b)
                    d.append(Algebra(POW, (b,e)))
                return mul_new(Algebra, d)
        return super(type(self), self).algebra_pow_number(Algebra, lhs, rhs, inplace)
NUMBER = NumberHead()




__all__ = ['AND', 'OR', 'NOT', 'IN', 'NOTIN', 'IS', 'ISNOT']

from .base import Head, UnaryHead, BinaryHead, NaryHead
from ..core import init_module, classes
init_module.import_heads()

class NotHead(UnaryHead):
    """
    NotHead represents unary boolean NOT operation,
    data is an expression operand.
    """
    op_symbol = 'not '
    def __repr__(self): return 'NOT'

    def new(self, cls, operand):
        assert type(operand) is cls,`type(operand), cls`
        head, data = operand.pair
        if head is NUMBER: return cls(NUMBER, not data)
        if head is NOT: return data
        if head is NE: return cls(EQ, data)
        if head is EQ: return cls(NE, data)
        if head is LT: return cls(GE, data)
        if head is GT: return cls(LE, data)
        if head is LE: return cls(GT, data)
        if head is GE: return cls(LT, data)
        return cls(NOT, operand)

    def reevaluate(self, cls, operand):
        return self.new(cls, operand)

class AndHead(NaryHead):
    """
    AndHead represents n-ary boolean AND operation,
    data is a n-tuple of expression operands.
    """
    op_symbol = ' and '
    def __repr__(self): return 'AND'

    def new(self, cls, operands):
        s = set([])
        for a in operands:
            assert type(a) is cls,`type(a)`
            head, data = a.pair
            if head is NUMBER:
                if not data:
                    return cls.false
            elif head is AND:
                s.update(data)
            elif head is NOT:
                if data in s:
                    return cls.false
                else:
                    s.add(a)
            else:
                if cls(NOT,a) in s:
                    return cls.false
                s.add(a)
        if not s:
            return cls.true
        if len(s)==1:
            return s.pop()
        return cls(AND, frozenset(s))

class OrHead(NaryHead):
    """
    AndHead represents n-ary boolean OR operation,
    data is a n-tuple of expression operands.
    """
    op_symbol = ' or '
    def __repr__(self): return 'OR'

    def new(self, cls, operands):
        s = set([])
        for a in operands:
            assert type(a) is cls,`type(a)`
            head, data = a.pair
            if head is NUMBER:
                if data:
                    return cls.true
            elif head is OR:
                s.update(data)
            elif head is NOT:
                if data in s:
                    return cls.true
                s.add(a)
            else:
                if cls(NOT, a) in s:
                    return cls.true
                s.add(a)
        if not s:
            return cls.false
        if len(s)==1:
            return s.pop()
        return cls(OR, frozenset(s))

class IsHead(BinaryHead):
    """
    IsHead represents binary boolean IS operation,
    data is an expression operand.
    """
    op_symbol = ' is '
    def __repr__(self): return 'IS'

class IsnotHead(BinaryHead):
    """
    IsHead represents binary boolean IS NOT operation,
    data is an expression operand.
    """
    op_symbol = ' is not '
    def __repr__(self): return 'ISNOT'

class InHead(BinaryHead):
    """
    InHead represents binary boolean IN operation,
    data is an expression operand.
    """
    op_symbol = ' in '

    def new(self, cls, (lhs, rhs)):
        return cls(self, (lhs, rhs))
    
    def __repr__(self): return 'IN'

    def walk(self, func, cls, data, target):
        lhs, rhs = data
        rcls = cls.get_operand_algebra(self, 1)
        rhs1 = rhs.head.walk(func, rcls, rhs.data, rhs)
        lcls = rhs1.get_element_algebra()
        lhs1 = lhs.head.walk(func, lcls, lhs.data, lhs)
        if lhs1 is lhs and rhs1 is rhs:
            return func(cls, data, target)
        r = self.new(cls, (lhs1, rhs1))
        return func(cls, r.head, r.data, r)

class NotinHead(BinaryHead):
    """
    NotinHead represents binary boolean NOT IN operation,
    data is an expression operand.
    """
    op_symbol = ' not in '
    def __repr__(self): return 'NOTIN'

NOT = NotHead()
OR = OrHead()
AND = AndHead()
IS = IsHead()
ISNOT = IsnotHead()
IN = InHead()
NOTIN = NotinHead()



'''
.. module:: parsers
   :platform: Linux
   :synopsis: Wrapper module for a number of parsing modules which are crucial for the functioning of PyMacLab. The first parser
              collects raw data (i.e. lines) from the DSGE model files, while the second parser carefully extracts useful information
              and then stores this by attaching it to the DSGE model instance.

.. moduleauthor:: Eric M. Scheffel <eric.scheffel@nottingham.edu.cn>


'''


# This file provides preprocess function and macros used by the
# mk_*.py scripts.
#
# Created by Pearu Peterson in Febuary 2008
#

def preprocess(source, globals_dict, tmp_cache=[1]):
    """ Preprocess source using macros from macros.py file and ones in
    globals_dict dictionary (the caller usually uses globals_dict=globals()).

    Macros start with `@` symbol and take 0 or more arguments separated
    with `;` symbol. Macros may use temporary symbols TMP, TMP0, ..., TMP5,
    that names are unique for every macro body.
    """
    g_dict = dict(globals())
    g_dict.update(globals_dict)
    result = []
    for line in source.splitlines():
        if line.lstrip().startswith('@'):
            prefix, rest = line.split('@',1)
            i = rest.index('(')
            name = rest[:i]
            tmp_cache[0] += 1
            d = {'TMP':'_tmp%s' % (tmp_cache[0])}
            for j in range(6):
                tmp_cache[0] += 1
                d['TMP'+str(j)] = '_tmp%s' % (tmp_cache[0])
            try:
                for arg in rest.strip()[i+1:-1].split(';'):
                    key, value = arg.split('=',1)
                    d[key.strip()] = value.strip()
            except Exception, msg:
                print '%s (while processing %r)' % (msg, line.lstrip())
                continue
            try:
                templ = eval(name, g_dict, {})
            except NameError, msg:
                templ = '@' + rest
                print 'NameError: %s (while processing %r)' % (msg, line.strip())
            else:
                if '@' in templ:
                    templ = preprocess(templ, globals_dict)
            result.append(prefix + '#' + rest)
            try:
                templ_d = templ % d
            except Exception, msg:
                print '%s (while processing %r)' % (msg, line.lstrip())
                #print d, `templ`
                continue
            for l in templ_d.splitlines():
                result.append(prefix + l)
        else:
            result.append(line)
    return '\n'.join(result)

NEWINSTANCE = '''\
%(OBJ)s = new(cls)
%(OBJ)s.head = %(HEAD)s
%(OBJ)s.data = %(DATA)s
'''
NEWINSTANCE = '''\
%(OBJ)s = new(cls, %(HEAD)s, %(DATA)s)
'''
NEWINSTANCE = '''\
%(OBJ)s = cls(%(HEAD)s, %(DATA)s)
'''
RETURN_NEW = '''\
@NEWINSTANCE(OBJ=%(TMP)s; HEAD=%(HEAD)s; DATA=%(DATA)s)
return %(TMP)s
'''

IF_CHECK_INT = 'if %(T)s is int or %(T)s is long:'
ELIF_CHECK_INT = 'elif %(T)s is int or %(T)s is long:'
IF_CHECK_REAL = 'if %(T)s is int or %(T)s is long or %(T)s is mpq or %(T)s is float or %(T)s is mpf:'
IF_CHECK_COMPLEX = 'if %(T)s is mpqc or %(T)s is complex or %(T)s is mpc:'

ELIF_CHECK_NUMBER = 'elif %(T)s is int or %(T)s is long or %(T)s is mpq or %(T)s is float or %(T)s is mpf or %(T)s is mpqc or %(T)s is mpc or %(T)s is complex:'

IF_CHECK_INT = 'if %(T)s in inttypes_set:'
ELIF_CHECK_INT = 'elif %(T)s in inttypes_set:'
IF_CHECK_REAL = 'if %(T)s in realtypes_set:'
IF_CHECK_COMPLEX = 'if %(T)s in complextypes_set:'
ELIF_CHECK_NUMBER = 'elif %(T)s in numbertypes_set:'

IF_CHECK_FLOAT = 'if %(T)s is float:'

ADD_TERM_VALUE_DICT='''\
%(TMP)s = %(DICT_GET)s(%(TERM)s)
if %(TMP)s is None:
    %(DICT)s[%(TERM)s] = %(USIGN)s %(VALUE)s
else:
    %(TMP)s = %(TMP)s %(SIGN)s %(VALUE)s
    if %(TMP)s:
        %(DICT)s[%(TERM)s] = %(TMP)s
    else:
        del %(DICT)s[%(TERM)s]
'''

MUL_FACTOR_VALUE_DICT='''\
%(TMP)s = %(DICT_GET)s(%(FACTOR)s)
if %(TMP)s is None:
    %(DICT)s[%(FACTOR)s] = %(USIGN)s %(VALUE)s
else:
    %(TMP)s = %(TMP)s %(SIGN)s %(VALUE)s
    if type(%(TMP)s) is cls and %(TMP)s.head is NUMBER:
        %(TMP)s = %(TMP)s.data
    if %(TMP)s:
        if %(FACTOR)s.head is NUMBER:
            del %(DICT)s[%(FACTOR)s]
            z, sym = try_power(%(FACTOR)s.data, %(TMP)s)
            if sym:
                for t1, c1 in sym:
                    @NEWINSTANCE(OBJ=tt; HEAD=NUMBER; DATA=t1)
                    @ADD_TERM_VALUE_DICT(DICT=%(DICT)s; DICT_GET=%(DICT_GET)s; TERM=tt; VALUE=c1; SIGN=+; USIGN=)
            %(NUMBER)s = %(NUMBER)s * z
        else:
            %(DICT)s[%(FACTOR)s] = %(TMP)s
    else:
        del %(DICT)s[%(FACTOR)s]
'''




#
# Created January 2008 by Pearu Peterson
#
""" Provides Calculus class.
"""
__docformat__ = "restructuredtext"
__all__ = ['Calculus', 'I']

from ..core import classes, defined_functions
from ..utils import TERMS, str_PRODUCT, SYMBOL, NUMBER
from ..heads import APPLY, CALLABLE, TERM_COEFF, TERM_COEFF_DICT
from ..heads import BASE_EXP_DICT as FACTORS
from ..basealgebra import Algebra, Verbatim
from ..ring import CommutativeRing

from ..arithmetic.numbers import normalized_fraction, mpq, mpf, mpc, mpqc, try_power

from ..arithmetic import mpmath, setdps
from ..arithmetic.evalf import evalf

algebra_numbers = (int, long, mpq, mpqc, mpf, mpc)
convertible_numbers = algebra_numbers + (float, complex)

float_one = mpf('1.0')

class Calculus(CommutativeRing):
    """ Represents an element of a symbolic algebra.

    The set of a symbolic algebra is a set of expressions.
    """

    _hash = None

    coefftypes = algebra_numbers
    coefftypes_set = frozenset(coefftypes)
    exptypes = algebra_numbers
    exptypes_set = frozenset(exptypes)

    @classmethod
    def get_function_algebra(cls):
        return classes.CalculusFunctionRing

    def __abs__(self):
        head, data = self.pair
        if head is NUMBER:
            return type(self)(NUMBER, abs(data))
        raise NotImplementedError('abs(%r)' % (self))

    def __call__(self, *args, **kws):
        head, data = self.pair
        if head is CALLABLE:
            return data(*args, **kws)
        raise TypeError(`self` + ' is not callable')

    def __lt__(self, other):
        if type(other) is type(self):
            return self.data < other.data
        return self.data < other
    def __le__(self, other):
        if type(other) is type(self):
            return self.data <= other.data
        return self.data <= other
    def __gt__(self, other):
        if type(other) is type(self):
            return self.data > other.data
        return self.data > other
    def __ge__(self, other):
        if type(other) is type(self):
            return self.data >= other.data
        return self.data >= other
    def __ne__(self, other):
        return not (self == other)
    
    def _sympy_(self):
        import sympy as m
        head, data = self.pair
        if head is SYMBOL:
            return m.Symbol(data)
        if head is NUMBER:
            return m.Number(data)
        if head is TERMS:
            return m.Add(*self.as_Add_args())
        if head is FACTORS:
            return m.Mul(*self.as_Mul_args())
        raise NotImplemented(`self`)

    def as_algebra(self, cls, typeerror=True):
        """ Convert algebra to another algebra.
        """
        if cls is Verbatim:
            return self.as_verbatim()
        if cls is classes.Unit:
            return cls(NUMBER, self)
        if issubclass(cls, PolynomialRing):
            return self.as_polynom(cls)
        return self.as_verbatim().as_algebra(cls)

    @classmethod
    def get_predefined_symbols(cls, name):
        if name=='I':
            return I
        if name=='D':
            return classes.DFactory()
        return getattr(defined_functions, name, None)

    @classmethod
    def get_defined_function(cls, name):
        func = getattr(defined_functions, name, None)
        if func is None:
            func = getattr(defined_functions, name.title(), None)
        return func
    
    @classmethod
    def convert_coefficient(cls, obj, typeerror=True):
        """ Convert obj to coefficient algebra.
        """
        if isinstance(obj, float):
            return mpf(obj)
        if isinstance(obj, complex):
            return mpc(obj.real, obj.imag)
        if isinstance(obj, algebra_numbers):
            return obj
        if isinstance(obj, cls):
            head, data = obj.pair
            if head is NUMBER:
                return data
        if typeerror:
            raise TypeError('%s.convert_coefficient: failed to convert %s instance'\
                            ' to coefficient algebra, expected int|long object'\
                            % (cls.__name__, obj.__class__.__name__))
        else:
            return NotImplemented

    @classmethod
    def convert_exponent(cls, obj, typeerror=True):
        """ Convert obj to exponent algebra.
        """
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, algebra_numbers):
            return obj
        if isinstance(obj, float):
            return mpf(obj)
        if isinstance(obj, complex):
            return mpc(obj.real, obj.imag)

        # parse algebra expression from string:
        if isinstance(obj, (str, unicode, Verbatim)):
            return Verbatim.convert(obj).as_algebra(cls)

        # convert from another algebra:
        if isinstance(obj, Algebra):
            return obj.as_algebra(cls)

        if typeerror:
            raise TypeError('%s.convert_exponent: failed to convert %s instance'\
                            ' to exponent algebra, expected int|long object'\
                            % (cls.__name__, obj.__class__.__name__))
        else:
            return NotImplemented
    
    @classmethod
    def Number(cls, num, denom=None):
        if denom is None:
            return cls(NUMBER, num)
        return cls(NUMBER, normalized_fraction(num, denom))

    @classmethod
    def Log(cls, arg, base=None):
        log = defined_functions.Log
        if base is None:
            return log(arg)
        return log(arg)/log(base)

    @classmethod
    def Exp(cls, arg):
        return defined_functions.Exp(arg)

    @classmethod
    def Mod(cls, x, y):
        return defined_functions.Mod(x, y)

    def evalf(self, n=None):
        if n is not None:
            setdps(n)
        head, data = self.pair
        def evalf(cls, head, data, target):
            if head is NUMBER:
                data = data * float_one
            if head is SYMBOL:
                if hasattr(data, 'evalf'):
                    return cls(data.evalf(n))
            if head is APPLY:
                func, args = data
                name = str(func).lower()
                if hasattr(mpmath, name):
                    f = getattr(mpmath, name)
                    l = []
                    for a in args:
                        h, d = a.pair
                        if h is NUMBER:
                            l.append(d)
                        else:
                            l = None
                            break
                    if l is not None:
                        return cls(f(*l))
            if head is TERM_COEFF:
                term, coeff = data
                coeff = coeff * float_one
                return cls(head, (term, coeff))
            if head is TERM_COEFF_DICT:
                new_data = {}
                for term, coeff in data.iteritems():
                    new_data[term] = coeff * float_one
                return cls(head, new_data)
            return target
        return head.walk(evalf, type(self), data, self)
        
        if n:
            setdps(n)
        head, data = self.pair
        if head is NUMBER:
            return self.Number(data * float_one)
        if head is SYMBOL:
            r = getattr(data, 'evalf', lambda p: NotImplemented)(n)
            if r is not NotImplemented:
                return self.Number(r)
            return self
        if head is APPLY:
            func = data[0]
            h, func_data = func.pair
            assert h is CALLABLE, `self, func` # todo: support symbolic functions
            args = data[1]
            assert len (args)==1,`self, args` # todo: support multivariate functions
            v = args[0].evalf(n)
            h, d = v.pair
            if h is NUMBER:
                return self.Number(getattr(mpmath, func_data.__name__.lower())(d))
            else:
                return type(self)(APPLY, (func, (v,)))
        convert = self.convert
        return self.func(*[convert(a).evalf(n) for a in self.args])

    def get_direction(self):
        head, data = self.pair
        if head is NUMBER:
            if isinstance(data, (int, long)):
                return data
            return getattr(data, 'get_direction', lambda : NotImplemented)()
        if head is TERMS:
            if len(data)==1:
                t, c = data.items()[0]
                r = t.get_direction()
                if r is not NotImplemented:
                    return r * c
        if head is FACTORS:
            direction = 1
            cls = type(self)
            for t,c in self.data.iteritems():
                d = t.get_direction()
                if d is NotImplemented:
                    return d
                if not isinstance(c, (int, long)):
                    return NotImplemented
                d = self.Pow(cls.convert(d), c).get_direction()
                if d is NotImplemented:
                    return d
                direction *= d
            return direction
        return getattr(data, 'get_direction', lambda : NotImplemented)()

    @property
    def is_bounded(self):
        head, data = self.pair
        if head is NUMBER:
            if isinstance(data, (int, long)):
                return True
            return getattr(data, 'is_bounded', None)
        if head is SYMBOL:
            return getattr(data, 'is_bounded', None)
        if head is TERMS:
            for t, c in data.iteritems():
                b = t.is_bounded
                if not b:
                    return b
                if isinstance(c, (int, long)):
                    continue
                b = getattr(c, 'is_bounded', None)
                if not b:
                    return b
            return True
        return

    def as_polynom(self, ring_cls=None):
        """ Convert expression to an element of polynomial ring.
        
        If the polynomial ring is not given then it will be created.

        For example,

          >>> x,y,z = map(Symbol,'xyz')
          >>> (x+2*y+3*z).as_polynom() + y*z
          PolynomialRing[(x, y, z), Calculus]('2*x + y*z + y + 3*z')
          >>> P = PolynomialRing[x,y]
          >>> (x+2*y+3*z).as_polynom(P) + y*z
          PolynomialRing[(x, y), Calculus]('x + (2 + z)*y + 3*z')

        """
        if ring_cls is None:
            data, variables = self.to_polynomial_data()
            return PolynomialRing[tuple(variables)].convert(data)
        else:
            data, variables = self.to_polynomial_data(ring_cls.variables, True)
            return ring_cls.convert(data)

    def __divmod__(self, other):
        if isinstance(other, Calculus):
            lhs = self.as_polynom()
            rhs = other.as_polynom(type(lhs))
            return divmod(lhs, rhs)
        return NotImplemented

    def __float__(self):
        head, data = self.pair
        if head is NUMBER:
            return float(data)
        head, data = self.evalf().pair
        if head is NUMBER:
            return float(data)
        raise TypeError('Cannot convert %r to float' % (self))

    def __or__(self, other):
        if type(other) is tuple:
            return self.subs(*other)
        return self.subs(other)

classes.Calculus = Calculus

# as a result, x<y will return Logic(LT, (x, y)):
Calculus.enable_symbolic_comparison('inequality')

one = Calculus.Number(1)
zero = Calculus.Number(0)
Calculus.one = one
Calculus.zero = zero

I = Calculus.Number(mpqc(0,1))

from ..polynomials.algebra import PolynomialRing, AdditiveTuple
from .infinity import CalculusInfinity




from sympycore.calculus import oo, moo, zoo, undefined, I, Calculus, pi

zero = Calculus.Number(0)
one = Calculus.Number(1)
mone = Calculus.Number(-1)
two = Calculus.Number(2)
mtwo = Calculus.Number(-2)
half = Calculus.Number(1)/2
mhalf = Calculus.Number(-1)/2
onehalf = Calculus.Number(3)/2
monehalf = Calculus.Number(-3)/2
fhalf = one/2.0

x,y,z = map(Calculus.Symbol, 'xyz')

def test_oo_product():
    assert str(oo * (x+y)) in ['oo*(x + y)', 'oo*(y + x)'], str(oo * (x+y))
    assert str(oo * (1+x+y)) in ['oo*(1 + x + y)', 'oo*(y + 1 + x)',
                                 'oo*(x + 1 + y)', 'oo*(y + x + 1)'],  str(oo * (1+x+y))
    assert str(oo * (2*x))=='oo*(x)'
    assert str(oo / (2/x))=='oo*(x)', `str(oo/(2/x))`
    assert str(oo / (2*x)) in ['oo*(x**(-1))','oo*(1/x)'],str(oo / (2*x))
    
def test_oo_sum():
    assert str(oo + (x+y)) in ['oo + (x + y)','oo + (y + x)'], str(oo + (x+y))
    assert str(oo + (2+x+y)) in ['oo + (2 + x + y)', 'oo + (y + 2 + x)',
                                 'oo + (x + 2 + y)', 'oo + (y + x + 2)'], str(oo + (2+x+y))
    assert str((2+x+y) + oo) in ['oo + (2 + x + y)', 'oo + (y + 2 + x)',
                                 'oo + (x + 2 + y)', 'oo + (y + x + 2)'], str((2+x+y) + oo)
    assert str(oo - (2+x+y)) in ['oo + (-2 - x - y)', 'oo + (-y - 2 - x)',
                                 'oo + (-x - 2 - y)', 'oo + (-y - x - 2)'], str(oo - (2+x+y))

    assert str((oo+y) + (x+y)) in ['oo + (y) + (x + y)', 'oo + (y) + (y + x)'], str((oo+y) + (x+y))
    assert str((x+y) + (oo+y)) in ['oo + (y) + (x + y)','oo + (y) + (y + x)'],  str((x+y) + (oo+y))
    assert str((oo+y) + (oo+x))=='oo + (y) + (oo + (x))', str((oo+y) + (oo+x))

    s = x+y
    s += oo
    assert str(s) in ['oo + (x + y)', 'oo + (y + x)'], str(s)

    assert str(Calculus.Add(oo, x))=='oo + (x)'
    assert str(Calculus.Add(x, oo))=='oo + (x)'
    assert str(Calculus.Add(y, oo, x))=='oo + (y) + (x)'
    assert str(Calculus.Add(y, oo, x, oo)) in ['oo + (y) + (oo + (x))', 'oo + (y) + (x) + (oo)'], str(Calculus.Add(y, oo, x, oo))

    s = x+y
    s -= -oo
    assert str(s) in ['oo + (x + y)', 'oo + (y + x)']

def test_oo_symbol():
    assert str(oo + x)=='oo + (x)'
    assert str(oo - x)=='oo + (-x)'
    assert str(x - oo)=='-oo + (x)'
    assert str(oo * x)=='oo*(x)'
    assert str(oo / x) in ['oo*(x**(-1))','oo*(1/x)'], str(oo / x)
    assert str(x/oo)=='0'
    assert str(oo ** x) in ['(oo)**x', 'oo**x'],str(oo**x)
    assert str(x**oo) in ['x**(oo)', 'x**oo'], str(x**oo)

    assert str(oo + pi)=='oo', str(oo+pi)
    assert str(oo - pi)=='oo'
    assert str(oo * pi)=='oo'
    assert str(oo / pi)=='oo'
    #assert str(oo ** pi)=='oo'
    assert str(pi - oo)=='-oo'
    assert str(pi/oo)=='0'
    #assert str(pi**oo)=='oo'

    assert oo + x == x + oo,`oo+x, x+oo`
    assert oo * x == x * oo
    assert oo + pi == pi + oo
    assert oo * pi == pi * oo

def test_moo_optable():
    assert moo + moo == moo
    assert moo + mtwo == moo
    assert moo + monehalf == moo
    assert moo + mone == moo
    assert moo + mhalf == moo
    assert moo + zero == moo
    assert moo + half == moo
    assert moo + fhalf == moo
    assert moo + one == moo
    assert moo + onehalf == moo
    assert moo + two == moo
    assert moo + oo == undefined
    assert moo + zoo == undefined
    assert moo + undefined == undefined

    assert moo - moo == undefined
    assert moo - mtwo == moo
    assert moo - monehalf == moo
    assert moo - mone == moo
    assert moo - mhalf == moo
    assert moo - zero == moo
    assert moo - half == moo
    assert moo - fhalf == moo
    assert moo - one == moo
    assert moo - onehalf == moo
    assert moo - two == moo
    assert moo - oo == moo
    assert moo - zoo == undefined
    assert moo - undefined == undefined

    assert moo * moo == oo
    assert moo * mtwo == oo
    assert moo * monehalf == oo
    assert moo * mone == oo
    assert moo * mhalf == oo
    assert moo * zero == undefined
    assert moo * half == moo
    assert moo * fhalf == moo
    assert moo * one == moo
    assert moo * onehalf == moo
    assert moo * two == moo
    assert moo * oo == moo
    assert moo * zoo == zoo
    assert moo * undefined == undefined

    assert moo / moo == undefined
    assert moo / mtwo == oo
    assert moo / monehalf == oo
    assert moo / mone == oo
    assert moo / mhalf == oo
    assert moo / zero == zoo
    assert moo / half == moo
    assert moo / fhalf == moo
    assert moo / one == moo
    assert moo / onehalf == moo
    assert moo / two == moo
    assert moo / oo == undefined
    assert moo / zoo == undefined
    assert moo / undefined == undefined

    assert moo ** moo == zero
    assert moo ** mtwo == zero
    assert moo ** monehalf == zero
    assert moo ** mone == zero
    assert moo ** mhalf == zero
    assert moo ** zero == one
    assert moo ** half == I*oo
    #assert moo ** fhalf == I*oo
    assert moo ** one == moo
    assert moo ** onehalf == -I*oo
    assert moo ** two == oo
    assert moo ** oo == zoo
    assert moo ** zoo == undefined
    assert moo ** undefined == undefined

def test_mtwo_optable():
    assert mtwo + moo == moo
    assert mtwo + oo == oo
    assert mtwo + undefined == undefined

    assert mtwo - moo == oo
    assert mtwo - oo == moo
    assert mtwo - undefined == undefined

    assert mtwo * moo == oo
    assert mtwo * oo == moo
    assert mtwo * zoo == zoo
    assert mtwo * undefined == undefined

    assert mtwo / moo == zero
    assert mhalf / zero == zoo
    assert mtwo / oo == zero
    assert mtwo / zoo == zero
    assert mtwo / undefined == undefined

    assert mtwo ** moo == zero
    assert mtwo ** oo == zoo
    assert mtwo ** zoo == undefined
    assert mtwo ** undefined == undefined

def test_monehalf_optable():
    assert monehalf + moo == moo
    assert monehalf + oo == oo
    assert monehalf + zoo == zoo
    assert monehalf + undefined == undefined

    assert monehalf - moo == oo
    assert monehalf - oo == moo
    assert monehalf - zoo == zoo
    assert monehalf - undefined == undefined

    assert monehalf * moo == oo
    assert monehalf * oo == moo
    assert monehalf * zoo == zoo
    assert monehalf * undefined == undefined

    assert monehalf / moo == zero
    assert mhalf / zero == zoo
    assert monehalf / oo == zero
    assert monehalf / zoo == zero
    assert monehalf / undefined == undefined

    assert monehalf ** moo == zero
    assert monehalf ** oo == zoo
    assert monehalf ** zoo == undefined
    assert monehalf ** undefined == undefined

def test_mone_optable():
    assert mone + moo == moo
    assert mone + oo == oo
    assert mone + undefined == undefined

    assert mone - moo == oo
    assert mone - oo == moo
    assert mone - undefined == undefined

    assert mone * moo == oo
    assert mone * oo == moo
    assert mone * undefined == undefined

    assert mone / moo == zero
    assert mhalf / zero == zoo
    assert mone / oo == zero
    assert mone / undefined == undefined

    assert mone ** moo == undefined
    assert mone ** oo == undefined
    assert mone ** undefined == undefined

def test_mhalf_optable():
    assert mhalf + moo == moo
    assert mhalf + oo == oo
    assert mhalf + zoo == zoo
    assert mhalf + undefined == undefined

    assert mhalf - moo == oo
    assert mhalf - oo == moo
    assert mhalf - zoo == zoo
    assert mhalf - undefined == undefined

    assert mhalf * moo == oo
    assert mhalf * oo == moo
    assert mhalf * zoo == zoo
    assert mhalf * undefined == undefined

    assert mhalf / moo == zero
    assert mhalf / zero == zoo
    assert mhalf / oo == zero
    assert mhalf / zoo == zero
    assert mhalf / undefined == undefined

    assert mhalf ** moo == zoo
    assert mhalf ** oo == zero
    assert mhalf ** zoo == undefined
    assert mhalf ** undefined == undefined

def test_zero_optable():
    assert zero + moo == moo
    assert zero + oo == oo
    assert zero + zoo == zoo
    assert zero + undefined == undefined

    assert zero - moo == oo
    assert zero - oo == moo
    assert zero - zoo == zoo
    assert zero - undefined == undefined

    assert zero * moo == undefined
    assert zero * oo == undefined
    assert zero * zoo == undefined
    assert zero * undefined == undefined

    assert zero / moo == zero
    assert zero / zero == undefined
    assert zero / oo == zero
    assert zero / zoo == zero
    assert zero / undefined == undefined

    assert zero ** moo == zoo
    assert zero ** oo == zero
    assert zero ** zoo == undefined
    assert zero ** undefined == undefined

def test_half_optable():
    assert half + moo == moo
    assert half + oo == oo
    assert half + zoo == zoo
    assert half + undefined == undefined

    assert half - moo == oo
    assert half - oo == moo
    assert half - zoo == zoo
    assert half - undefined == undefined

    assert half * moo == moo
    assert half * oo == oo
    assert half * zoo == zoo
    assert half * undefined == undefined

    assert half / moo == zero
    assert half / zero == zoo
    assert half / oo == zero
    assert half / zoo == zero
    assert half / undefined == undefined

    assert half ** moo == zoo
    assert half ** oo == zero
    assert half ** zoo == undefined
    assert half ** undefined == undefined

def test_fhalf_optable():
    assert fhalf + moo == moo
    assert fhalf + oo == oo
    assert fhalf + zoo == zoo
    assert fhalf + undefined == undefined

    assert fhalf - moo == oo
    assert fhalf - oo == moo
    assert fhalf - zoo == zoo
    assert fhalf - undefined == undefined

    assert fhalf * moo == moo
    assert fhalf * oo == oo
    assert fhalf * zoo == zoo
    assert fhalf * undefined == undefined

    assert fhalf / moo == zero
    assert fhalf / zero == zoo
    assert fhalf / oo == zero
    assert fhalf / zoo == zero
    assert fhalf / undefined == undefined

    assert fhalf ** moo == zoo
    assert fhalf ** oo == zero
    assert fhalf ** zoo == undefined
    assert fhalf ** undefined == undefined

def test_one_optable():
    assert one + moo == moo
    assert one + oo == oo
    assert one + zoo == zoo
    assert one + undefined == undefined

    assert one - moo == oo
    assert one - oo == moo
    assert one - zoo == zoo
    assert one - undefined == undefined

    assert one * moo == moo
    assert one * oo == oo
    assert one * zoo == zoo
    assert one * undefined == undefined

    assert one / moo == zero
    assert one / zero == zoo
    assert one / oo == zero
    assert one / zoo == zero
    assert one / undefined == undefined

    assert one ** moo == one
    assert one ** oo == one
    assert one ** zoo == one
    assert one ** undefined == one

def test_onehalf_optable():
    assert onehalf + moo == moo
    assert onehalf + oo == oo
    assert onehalf + zoo == zoo
    assert onehalf + undefined == undefined

    assert onehalf - moo == oo
    assert onehalf - oo == moo
    assert onehalf - zoo == zoo
    assert onehalf - undefined == undefined

    assert onehalf * moo == moo
    assert onehalf * oo == oo
    assert onehalf * zoo == zoo
    assert onehalf * undefined == undefined

    assert onehalf / moo == zero
    assert onehalf / zero == zoo
    assert onehalf / oo == zero
    assert onehalf / zoo == zero
    assert onehalf / undefined == undefined

    assert onehalf ** moo == zero
    assert onehalf ** oo == oo
    assert onehalf ** zoo == undefined
    assert onehalf ** undefined == undefined

def test_two_optable():
    assert two + moo == moo
    assert two + oo == oo
    assert two + zoo == zoo
    assert two + undefined == undefined

    assert two - moo == oo
    assert two - oo == moo
    assert two - zoo == zoo
    assert two - undefined == undefined

    assert two * moo == moo
    assert two * oo == oo
    assert two * zoo == zoo
    assert two * undefined == undefined

    assert two / moo == zero
    assert two / zero == zoo, `two/zero`
    assert two / oo == zero
    assert two / zoo == zero
    assert two / undefined == undefined

    assert two ** moo == zero
    assert two ** oo == oo
    assert two ** zoo == undefined
    assert two ** undefined == undefined

def test_oo_optable():
    assert oo + moo == undefined
    assert oo + mtwo == oo
    assert oo + monehalf == oo
    assert oo + mone == oo
    assert oo + mhalf == oo
    assert oo + zero == oo
    assert oo + half == oo
    assert oo + fhalf == oo
    assert oo + one == oo
    assert oo + onehalf == oo
    assert oo + two == oo
    assert oo + oo == oo
    assert oo + zoo == undefined
    assert oo + undefined == undefined

    assert oo - moo == oo
    assert oo - mtwo == oo
    assert oo - monehalf == oo
    assert oo - mone == oo
    assert oo - mhalf == oo
    assert oo - zero == oo
    assert oo - half == oo
    assert oo - fhalf == oo
    assert oo - one == oo
    assert oo - onehalf == oo
    assert oo - two == oo
    assert oo - oo == undefined
    assert oo - zoo == undefined
    assert oo - undefined == undefined

    assert oo * moo == moo
    assert oo * mtwo == moo
    assert oo * monehalf == moo
    assert oo * mone == moo
    assert oo * mhalf == moo
    assert oo * zero == undefined
    assert oo * half == oo
    assert oo * fhalf == oo
    assert oo * one == oo
    assert oo * onehalf == oo
    assert oo * two == oo
    assert oo * oo == oo
    assert oo * zoo == zoo
    assert oo * undefined == undefined

    assert oo / moo == undefined
    assert oo / mtwo == moo
    assert oo / monehalf == moo
    assert oo / mone == moo
    assert oo / mhalf == moo
    assert oo / zero == zoo
    assert oo / half == oo
    assert oo / fhalf == oo
    assert oo / one == oo
    assert oo / onehalf == oo
    assert oo / two == oo
    assert oo / oo == undefined
    assert oo / zoo == undefined
    assert oo / undefined == undefined

    assert oo ** moo == zero
    assert oo ** mtwo == zero
    assert oo ** monehalf == zero
    assert oo ** mone == zero
    assert oo ** mhalf == zero
    assert oo ** zero == one
    assert oo ** half == oo
    assert oo ** fhalf == oo
    assert oo ** one == oo
    assert oo ** onehalf == oo
    assert oo ** two == oo
    assert oo ** oo == oo
    assert oo ** zoo == undefined
    assert oo ** undefined == undefined

def test_zoo_optable():
    assert zoo + moo == undefined
    assert zoo + mtwo == zoo
    assert zoo + monehalf == zoo
    assert zoo + mone == zoo
    assert zoo + mhalf == zoo
    assert zoo + zero == zoo
    assert zoo + half == zoo
    assert zoo + fhalf == zoo
    assert zoo + one == zoo
    assert zoo + onehalf == zoo
    assert zoo + two == zoo
    assert zoo + oo == undefined
    assert zoo + zoo == undefined
    assert zoo + undefined == undefined

    assert zoo - moo == undefined
    assert zoo - mtwo == zoo
    assert zoo - monehalf == zoo
    assert zoo - mone == zoo
    assert zoo - mhalf == zoo
    assert zoo - zero == zoo
    assert zoo - half == zoo
    assert zoo - fhalf == zoo
    assert zoo - one == zoo
    assert zoo - onehalf == zoo
    assert zoo - two == zoo
    assert zoo - oo == undefined
    assert zoo - zoo == undefined
    assert zoo - undefined == undefined

    assert zoo * moo == zoo
    assert zoo * mtwo == zoo
    assert zoo * monehalf == zoo
    assert zoo * mone == zoo
    assert zoo * mhalf == zoo
    assert zoo * zero == undefined
    assert zoo * half == zoo
    assert zoo * fhalf == zoo
    assert zoo * one == zoo
    assert zoo * onehalf == zoo
    assert zoo * two == zoo
    assert zoo * oo == zoo
    assert zoo * zoo == zoo
    assert zoo * undefined == undefined

    assert zoo / moo == undefined
    assert zoo / mtwo == zoo
    assert zoo / monehalf == zoo
    assert zoo / mone == zoo
    assert zoo / mhalf == zoo
    assert zoo / zero == zoo
    assert zoo / half == zoo
    assert zoo / fhalf == zoo
    assert zoo / one == zoo
    assert zoo / onehalf == zoo
    assert zoo / two == zoo
    assert zoo / oo == undefined
    assert zoo / zoo == undefined
    assert zoo / undefined == undefined

    assert zoo ** moo == zero
    assert zoo ** mtwo == zero
    assert zoo ** monehalf == zero
    assert zoo ** mone == zero
    assert zoo ** mhalf == zero
    assert zoo ** zero == one
    assert zoo ** half == zoo
    assert zoo ** fhalf == zoo
    assert zoo ** one == zoo
    assert zoo ** onehalf == zoo
    assert zoo ** two == zoo
    assert zoo ** oo == zoo
    assert zoo ** zoo == undefined
    assert zoo ** undefined == undefined

def test_undefined_optable():
    assert undefined + moo == undefined
    assert undefined + mtwo == undefined
    assert undefined + monehalf == undefined
    assert undefined + mone == undefined
    assert undefined + mhalf == undefined
    assert undefined + zero == undefined
    assert undefined + half == undefined
    assert undefined + fhalf == undefined
    assert undefined + one == undefined
    assert undefined + onehalf == undefined
    assert undefined + two == undefined
    assert undefined + oo == undefined
    assert undefined + zoo == undefined
    assert undefined + undefined == undefined

    assert undefined - moo == undefined
    assert undefined - mtwo == undefined
    assert undefined - monehalf == undefined
    assert undefined - mone == undefined
    assert undefined - mhalf == undefined
    assert undefined - zero == undefined
    assert undefined - half == undefined
    assert undefined - fhalf == undefined
    assert undefined - one == undefined
    assert undefined - onehalf == undefined
    assert undefined - two == undefined
    assert undefined - oo == undefined
    assert undefined - zoo == undefined
    assert undefined - undefined == undefined

    assert undefined * moo == undefined
    assert undefined * mtwo == undefined
    assert undefined * monehalf == undefined
    assert undefined * mone == undefined
    assert undefined * mhalf == undefined
    assert undefined * zero == undefined
    assert undefined * half == undefined
    assert undefined * fhalf == undefined
    assert undefined * one == undefined
    assert undefined * onehalf == undefined
    assert undefined * two == undefined
    assert undefined * oo == undefined
    assert undefined * zoo == undefined
    assert undefined * undefined == undefined

    assert undefined / moo == undefined
    assert undefined / mtwo == undefined
    assert undefined / monehalf == undefined
    assert undefined / mone == undefined
    assert undefined / mhalf == undefined
    assert undefined / zero == undefined
    assert undefined / half == undefined
    assert undefined / fhalf == undefined
    assert undefined / one == undefined
    assert undefined / onehalf == undefined
    assert undefined / two == undefined
    assert undefined / oo == undefined
    assert undefined / zoo == undefined
    assert undefined / undefined == undefined

    assert undefined ** moo == undefined
    assert undefined ** mtwo == undefined
    assert undefined ** monehalf == undefined
    assert undefined ** mone == undefined
    assert undefined ** mhalf == undefined
    assert undefined ** zero == one
    assert undefined ** half == undefined
    assert undefined ** fhalf == undefined
    assert undefined ** one == undefined
    assert undefined ** onehalf == undefined
    assert undefined ** two == undefined
    assert undefined ** oo == undefined
    assert undefined ** zoo == undefined
    assert undefined ** undefined == undefined

def test_issue62():
    z = one / zoo
    assert z==0
    assert type(z) is Calculus, `type(z)`
